@@ -14,6 +14,7 @@ dbt.adapters.factory.get_adapter = lambda config: config.adapter
 import atexit
 import os
 import threading
+import uuid
 from collections import OrderedDict, UserDict
 from copy import copy
 from enum import Enum
@@ -148,6 +149,20 @@ class ConfigInterface:
         self.single_threaded = threads == 1
         self.quiet = True
 
+    @classmethod
+    def from_str(cls, arguments: str) -> "ConfigInterface":
+        import argparse
+        import shlex
+
+        parser = argparse.ArgumentParser()
+        args = parser.parse_args(shlex.split(arguments))
+        return cls(
+            threads=args.threads,
+            target=args.target,
+            profiles_dir=args.profiles_dir,
+            project_dir=args.project_dir,
+        )
+
 
 class YamlHandler(YAML):
     """A `ruamel.yaml` wrapper to handle dbt YAML files with sane defaults"""
@@ -242,7 +257,7 @@ class DbtProject:
             set_from_args(self.args, self.args)
             self.config = RuntimeConfig.from_args(self.args)
             if hasattr(self, "adapter"):
-                self.adapter.cleanup_all()
+                self.adapter.connections.cleanup_all()
             self.adapter = self._verify_connection(self.get_adapter())
             self.config.adapter = self.adapter
 
@@ -252,6 +267,10 @@ class DbtProject:
         self.dbt = project_parser.load()
         self.dbt.build_flat_graph()
         project_parser.save_macros_to_adapter(self.adapter)
+        self._sql_parser = None
+        self._macro_parser = None
+        self._sql_compiler = None
+        self._sql_runner = None
 
     @classmethod
     def from_args(cls, args: ConfigInterface) -> "DbtProject":
@@ -264,7 +283,7 @@ class DbtProject:
         )
 
     @property
-    def yaml_handler(self) -> SqlBlockParser:
+    def yaml_handler(self) -> YamlHandler:
         """A YAML handler for loading and dumping yaml files on disk"""
         if self._yaml_handler is None:
             self._yaml_handler = YamlHandler()
@@ -318,6 +337,19 @@ class DbtProject:
         else:
             return adapter
 
+    def adapter_probe(self) -> bool:
+        """Check adapter connection, useful for long running processes such as the server or workbench"""
+        if not hasattr(self, "adapter") or self.adapter is None:
+            return False
+        try:
+            with self.adapter.connection_named("osmosis-heartbeat"):
+                self.adapter.debug_query()
+        except Exception:
+            # TODO: we can decide to reinit the Adapter here
+            return False
+        logger().info("Heartbeat received for %s", self.project_name)
+        return True
+
     def fn_threaded_conn(self, fn: Callable[..., T], *args, **kwargs) -> Callable[..., T]:
         """Used for jobs which are intended to be submitted to a thread pool,
         the 'master' thread should always have an available connection for the duration of
@@ -372,10 +404,6 @@ class DbtProject:
         self.get_macro_function.cache_clear()
         self.get_columns.cache_clear()
         self.compile_sql.cache_clear()
-        self._sql_parser = None
-        self._macro_parser = None
-        self._sql_compiler = None
-        self._sql_runner = None
 
     @lru_cache(maxsize=10)
     def get_ref_node(self, target_model_name: str) -> MaybeNonSource:
@@ -456,9 +484,21 @@ class DbtProject:
         return self.execute_sql(compiled_sql or raw_sql)
 
     @lru_cache(maxsize=SQL_CACHE_SIZE)
-    def compile_sql(self, raw_sql: str) -> DbtAdapterCompilationResult:
-        """Creates a node with a `dbt.parser.sql` class. Compile generated node."""
-        return self.compile_node(self.get_server_node(raw_sql))
+    def compile_sql(self, raw_sql: str, retry: int = 3) -> DbtAdapterCompilationResult:
+        """Creates a node with `get_server_node` method. Compile generated node.
+        Has a retry built in because even uuidv4 cannot gaurantee uniqueness at the speed
+        in which we can call this function concurrently. A retry significantly increases the stability"""
+        temp_node_id = str(uuid.uuid4())
+        try:
+            node = self.compile_node(self.get_server_node(raw_sql, temp_node_id))
+        except Exception as exc:
+            if retry > 0:
+                return self.compile_sql(raw_sql, retry - 1)
+            raise exc
+        else:
+            return node
+        finally:
+            self._clear_node(temp_node_id)
 
     def compile_node(self, node: ManifestNode) -> DbtAdapterCompilationResult:
         """Compiles existing node."""
@@ -592,7 +632,7 @@ class DbtProjectContainer:
         if project is None:
             return
         project.clear_caches()
-        project.adapter.cleanup_all()
+        project.adapter.connections.cleanup_all()
         self._projects.pop(project_name)
         if self._default_project == project_name:
             if len(self) > 0:
@@ -603,7 +643,8 @@ class DbtProjectContainer:
     def drop_all_projects(self) -> None:
         """Drop all DbtProjectContainers"""
         self._default_project = None
-        self._projects.clear()
+        for project in self._projects:
+            self.drop_project(project)
 
     def reparse_all_projects(self) -> None:
         """Reparse all projects"""
@@ -928,7 +969,7 @@ class DbtYamlManager(DbtProject):
                             continue
                         # TODO: We avoid sources for complexity reasons but if we are opinionated, we don't have to
                         schema = self.assert_schema_has_no_sources(
-                            self.yaml.load(schema_file.current)
+                            self.yaml_handler.load(schema_file.current)
                         )
                         models_in_file: Iterable[Dict[str, Any]] = schema.get("models", [])
                         for documented_model in models_in_file:
@@ -984,22 +1025,24 @@ class DbtYamlManager(DbtProject):
                 if not self.dry_run:
                     target.parent.mkdir(exist_ok=True, parents=True)
                     target.touch()
-                    self.yaml.dump(structure.output, target)
+                    self.yaml_handler.dump(structure.output, target)
 
             else:
                 # Update File
                 logger().info(":toolbox: Updating schema file %s", target.name)
-                target_schema: Dict[str, Any] = self.yaml.load(target)
-                if "version" not in target_schema:
+                target_schema: Optional[Dict[str, Any]] = self.yaml_handler.load(target)
+                if not target_schema:
+                    target_schema = {"version": 2}
+                elif "version" not in target_schema:
                     target_schema["version"] = 2
                 target_schema.setdefault("models", []).extend(structure.output["models"])
                 if not self.dry_run:
-                    self.yaml.dump(target_schema, target)
+                    self.yaml_handler.dump(target_schema, target)
 
             # Clean superseded schema files
             for dir, models in structure.supersede.items():
                 preserved_models = []
-                raw_schema: Dict[str, Any] = self.yaml.load(dir)
+                raw_schema: Dict[str, Any] = self.yaml_handler.load(dir)
                 models_marked_for_superseding = set(models)
                 models_in_schema = set(map(lambda mdl: mdl["name"], raw_schema.get("models", [])))
                 non_superseded_models = models_in_schema - models_marked_for_superseding
@@ -1013,7 +1056,7 @@ class DbtYamlManager(DbtProject):
                             preserved_models.append(model)
                     raw_schema["models"] = preserved_models
                     if not self.dry_run:
-                        self.yaml.dump(raw_schema, dir)
+                        self.yaml_handler.dump(raw_schema, dir)
                     logger().info(
                         ":satellite: Model documentation migrated from %s to %s",
                         dir.name,
@@ -1166,7 +1209,7 @@ class DbtYamlManager(DbtProject):
                 n_cols_doc_inherited = 0
                 n_cols_removed = 0
                 if len(missing_columns) > 0 or len(undocumented_columns) or len(extra_columns) > 0:
-                    schema_file = self.yaml.load(schema_path.current)
+                    schema_file = self.yaml_handler.load(schema_path.current)
                     (
                         n_cols_added,
                         n_cols_doc_inherited,
@@ -1181,7 +1224,7 @@ class DbtYamlManager(DbtProject):
                     if n_cols_added + n_cols_doc_inherited + n_cols_removed > 0:
                         # Dump the mutated schema file back to the disk
                         if not self.dry_run:
-                            self.yaml.dump(schema_file, schema_path.current)
+                            self.yaml_handler.dump(schema_file, schema_path.current)
                         logger().info(":sparkles: Schema file updated")
 
                 # Print Audit Report
